from .receiver_dsp import cd_compensation
from .receiver_dsp import MatchedFilter
from .receiver_dsp import syncsignal
from .receiver_dsp import dual_pol_time_domain_lms_equalizer as dual_lms
from .receiver_dsp import dual_pol_time_domain_lms_equalizer_pll as dual_lms_pll
from .receiver_dsp import superscalar
from .receiver_metrics import biterror
from .receiver_metrics import estimate_snr_using_tx


'''
    This file contains optical instrument along the signal's propagation
'''
import numpy as np
from numpy.fft import fftfreq
from scipy.constants import h, c
from scipy.fftpack import fft, ifft
import tqdm
from ..Base.SignalInterface import QamSignal, Signal, WdmSignal
from ..Filter.designFilter import LowPassFilter
from typing import List
from scipy.special import erf
import warnings
import numba


class Multiplex(object):
    '''
        This class implements signal multiplex
    '''

    @staticmethod
    def mux_signal(signals: List[Signal], grid_size):
        '''

        :param signals: signals that will be
        :param grid_size: the grid size of the wdm signal
        :return:
        '''
        fs_in_fiber = []
        for signal in signals:
            fs_in_fiber.append(signal.fs_in_fiber)

        if np.any(np.diff(np.array(fs_in_fiber))):
            assert "the sampling frequence of each signal in fiber should be the same"

        absolute_frequences = []
        for signal in signals:
            absolute_frequences.append(signal.center_frequence)

        absolute_frequences = np.array(absolute_frequences)
        freq = np.min(absolute_frequences) + np.max(absolute_frequences)
        freq = freq / 2

        relative_frequence = np.array(absolute_frequences) - freq
        # 从左到右开始复用
        # 每个wdm信道的采样频率必须保持一致
        # 每个wdm信道的样本点个数也要保持一致,截掉长的采样序列的尾部

        sample_length = []
        for signal in signals:
            sample_length.append(signal.sample_number_in_fiber)

        if np.any(np.diff(sample_length)):
            warnings.warn("the number of sample in fiber are not the same, the longer will be "
                          "trancted")
            min_length = min(sample_length)

            for signal in signals:
                number_to_del = signal.sample_number_in_fiber - min_length
                if number_to_del == 0:
                    continue
                else:
                    signal.data_sample_in_fiber = signal.data_sample_in_fiber[:, 0: - number_to_del]

        sample_number = signals[0].sample_number_in_fiber
        fs = signals[0].fs_in_fiber
        t_array = np.arange(0, sample_number) * (1 / fs)

        # pol_number = signals[0].pol_number
        channel_number = len(signals)
        # sample_length = signals[0][:].shape[1]

        # samples = np.zeros((pol_number, channel_number, sample_length), dtype=np.complex128)
        wdm_data_sample = 0 + 0j

        for ch_index in tqdm.tqdm(range(channel_number), ascii=True):
            wdm_data_sample = wdm_data_sample + mux_signal2(signals[ch_index][:], t_array, relative_frequence[ch_index])

        # wdm_data_sample = mux_signal(samples, t_array, relative_frequence.reshape(-1, 1))

        wdm_signal = WdmSignal(signals, grid_size, wdm_data_sample)

        return wdm_signal


@numba.njit("complex128[:,:](complex128[:,:],float64[:],float64)", cache=True)
def mux_signal2(samples, t_array, relative_frequence):
    exp_factor = np.exp(1j * 2 * np.pi * relative_frequence * t_array)
    samples = samples * exp_factor
    return samples


@numba.njit("complex128[:,:](complex128[:,:,:], float64[:],float64[:,:])", cache=True)
def mux_signal(samples, t_array, relative_frequence):
    '''
        each row represent a channel
    '''
    pol_number = samples.shape[0]
    channel_number = samples.shape[1]
    wdm_signal_datasample = np.zeros((pol_number, samples.shape[2]), dtype=np.complex128)

    exp_factor = np.exp(1j * 2 * np.pi * relative_frequence * t_array)
    for pol_index in range(pol_number):
        wdm_signal_datasample[pol_index, :] = np.sum(samples[pol_index, :] * exp_factor, axis=0)
    wdm_signal_datasample = np.atleast_2d(wdm_signal_datasample)
    return wdm_signal_datasample


class Demultiplex(object):

    @staticmethod
    def demux_signal(wdm_signal, signal_index=None, roll_off=0.02, cls=QamSignal):
        '''

        This static function is used to demultiplex a wdm_signal, the signal_index of wdm_signal will be obtained, it
        will be shift to center_frequency, an ideal low pass filter will be used to remove other signal that is not
        interested in. Note CD should do first

        :param wdm_signal: the wdm signal to be demulitplexed
        :param signal_index: which signal want to get,return signal objectl,if None.all channel will be returned
        :return: a QamSignal
        '''
        frequences = wdm_signal.relative_frequences
        tarray = np.arange(0, wdm_signal[:].shape[1]) * (1 / wdm_signal.fs_in_fiber)

        # temp = wdm_signal[:]
        temp = np.zeros_like(wdm_signal[:])

        for i in range(temp.shape[0]):
            temp[i, :] = wdm_signal[i, :] * np.exp(-1j * 2 * np.pi * frequences[signal_index] * tarray)

        # center_frequence = wdm_signal.relative_frequence[signal_index]
        pos_freq = wdm_signal.signals[signal_index].symbol_rate / 2 * (1 + roll_off)
        neg_freq = -pos_freq
        # LowPassFilter.inplace_ideal_lowpass(temp,pos_freq,neg_freq,fs_in_fiber=wdm_signal.fs_in_fiber)

        signal = cls(symbol_rate=wdm_signal.signals[signal_index].symbol_rate,
                     sps=wdm_signal.signals[signal_index].sps,
                     sps_in_fiber=wdm_signal.signals[signal_index].sps_in_fiber,
                     is_dp=wdm_signal.signals[signal_index].is_dp, is_from_array=True, is_from_demux=True,
                     mf=wdm_signal.signals[signal_index].mf,
                     symbol_length=wdm_signal.signals[signal_index].symbol_length)

        signal.data_sample_in_fiber = temp
        #
        signal._symbol = wdm_signal.signals[signal_index]._symbol
        signal.message = wdm_signal.signals[signal_index].message
        # the symbol and msg should not be changed, it is the tx information
        signal.center_frequence = wdm_signal.center_frequence
        LowPassFilter.inplace_ideal_lowpass(signal, pos_freq, neg_freq)
        # signal.symbol = wdm_signal.signals[signal_index].symbol
        return signal


class Edfa:

    def __init__(self, gain_db, nf, is_ase=True, mode='ConstantGain', expected_power=0):
        '''

        :param gain_db:
        :param nf:
        :param is_ase: 是否添加ase噪声
        :param mode: ConstantGain or ConstantPower
        :param expected_power: 当mode为ConstantPoower  时候，此参数有效
        '''

        self.gain_db = gain_db
        self.nf = nf
        self.is_ase = is_ase
        self.mode = mode
        self.expected_power = expected_power

    def one_ase(self, signal, gain_lin=None):
        '''

        :param signal:
        :return:
        '''
        lamb = signal.center_wave_length
        if gain_lin is None:
            one_ase = (h * c / lamb) * (self.gain_lin * self.nf_lin - 1) / 2
        else:
            one_ase = (h * c / lamb) * (gain_lin * self.nf_lin - 1) / 2
        return one_ase

    @property
    def gain_lin(self):
        return 10 ** (self.gain_db / 10)

    @property
    def nf_lin(self):
        return 10 ** (self.nf / 10)

    def traverse(self, signal):
        if self.is_ase:
            noise = self.one_ase(signal) * signal.fs_in_fiber
            noise_sample = np.random.randn(*(signal[:].shape)) + 1j * np.random.randn(*(signal[:].shape))
            each_pol_power = noise

            noise_sample = np.sqrt(each_pol_power / 2) * noise_sample

        else:
            noise_sample = 0

        if self.mode == 'ConstantGain':
            signal[:] = np.sqrt(self.gain_lin) * signal[:] + noise_sample
            return

        if self.mode == 'ConstantPower':
            #             raise NotImplementedError("Not implemented")
            signal_power = np.mean(np.abs(signal.data_sample[0, :]) ** 2) + np.mean(
                np.abs(signal.data_sample[1, :]) ** 2)
            desired_power_linear = (10 ** (self.expected_power / 10)) / 1000
            linear_gain = desired_power_linear / signal_power
            #
            #
            noise = self.one_ase(signal, linear_gain) * signal.fs_in_fiber
            noise_sample = np.random.randn(*(signal[:].shape)) + 1j * np.random.randn(*(signal[:].shape))
            noise_sample = np.sqrt(noise / 2) * noise_sample
            signal[:] = np.sqrt(linear_gain) * signal[:] + noise_sample

        return signal

    def __call__(self, signal):
        self.traverse(signal)
        return signal

    def __str__(self):

        string = f"Model is {self.mode}\n" \
            f"Gain is {self.gain_db} db\n" \
            f"ase is {self.is_ase}\n" \
            f"noise figure is {self.nf}"
        return string

    def __repr__(self):
        return self.__str__()


class WSS(object):

    def __init__(self, frequency_offset, bandwidth, oft):

        '''

        :param frequency_offset: value away from center [GHz]
        :param bandwidth: 3-db Bandwidth [Ghz]
        :param oft:GHZ
        '''
        self.frequency_offset = frequency_offset
        self.bandwidth = bandwidth
        self.otf = oft
        self.H = None
        self.freq = None

    def traverse(self, signal):

        sample = np.zeros_like(signal[:])
        for i in range(sample.shape[0]):
            sample[i, :] = signal[i, :]

        freq = fftfreq(len(sample[0, :]), 1 / signal.fs_in_fiber)
        freq = freq / 1e9
        self.freq = freq
        self.__get_transfer_function(freq)

        for i in range(sample.shape[0]):
            sample[i, :] = ifft(fft(sample[i, :]) * self.H)

        return sample

    def __call__(self, signal):
        sample = self.traverse(signal)
        signal[:] = sample
        return signal

    def __get_transfer_function(self, freq_vector):
        delta = self.otf / 2 / np.sqrt(2 * np.log(2))

        H = 0.5 * delta * np.sqrt(2 * np.pi) * (
                erf((self.bandwidth / 2 - (freq_vector - self.frequency_offset)) / np.sqrt(2) / delta) - erf(
            (-self.bandwidth / 2 - (freq_vector - self.frequency_offset)) / np.sqrt(2) / delta))

        H = H / np.max(H)

        self.H = H

    def plot_transfer_function(self, freq=None):
        import matplotlib.pyplot as plt
        if self.H is None:
            self.__get_transfer_function(freq)
            self.freq = freq
        index = self.H > 0.001
        plt.figure(figsize=(20, 6))
        plt.subplot(121)
        plt.scatter(self.freq[index], np.abs(self.H[index]), color='b', marker='o')
        plt.xlabel('GHz')
        plt.ylabel('Amplitude')
        plt.title("without log")
        plt.subplot(122)
        plt.scatter(self.freq[index], 10 * np.log10(np.abs(self.H[index])), color='b', marker='o')
        plt.xlabel('GHz')
        plt.ylabel('Amplitude')
        plt.title("with log")
        plt.show()

    def __str__(self):

        string = f'the center_frequency is {0 + self.frequency_offset}[GHZ] \t\n' \
            f'the 3-db bandwidth is {self.bandwidth}[GHz]\t\n' \
            f'the otf is {self.otf} [GHz] \t\n'
        return string

    def __repr__(self):
        return self.__str__()


class MZM(object):
    pass


class IQ(object):
    pass


class LaserSource(object):
    pass



import os

import numba
import numpy as np
from numba import prange
from scipy.fftpack import fft, ifft, fftfreq
from scipy.signal import correlate, lfilter
from scipy.signal import fftconvolve

from .dsp_tools import cal_scaling_factor_qam
from .dsp_tools import cal_symbols_qam
from .dsp_tools import decision
from .dsp_tools import segment_axis
from .dsp_tools import exp_decision
from ..Base.SignalInterface import Signal
from ..Instrument.ElectricInstrument import PulseShaping


class MatchedFilter(object):

    def __init__(self, roll_off=0.02, span=1024, sps=4):
        self.h = PulseShaping.rcosdesign(int(span * sps), roll_off, 1, int(sps))
        self.delay = int(span / 2 * sps)

    def match_filter(self, signal):
        x = signal.data_sample_in_fiber[0, :]
        y = signal.data_sample_in_fiber[1, :]
        x = fftconvolve(x, self.h)
        y = fftconvolve(y, self.h)

        x = np.roll(x, -self.delay)
        y = np.roll(y, -self.delay)
        x = x[:signal.sample_number_in_fiber]
        y = y[:signal.sample_number_in_fiber]
        return x, y

    def inplace_match_filter(self, signal):
        x, y = self.match_filter(signal)
        signal.data_sample_in_fiber = np.array([x, y])
        return signal

    def __call__(self, signal):
        self.inplace_match_filter(signal)

        return signal


def syncsignal(symbol_tx, sample_rx, sps):
    '''

        :param symbol_tx: 发送符号
        :param sample_rx: 接收符号，会相对于发送符号而言存在滞后
        :param sps: samples per symbol
        :return: 收端符号移位之后的结果

        # 不会改变原信号

    '''
    assert sample_rx.ndim == 1
    assert symbol_tx.ndim == 1
    assert len(sample_rx) >= len(symbol_tx)
    symbol_tx = np.atleast_2d(symbol_tx)[0, :]
    sample_rx = np.atleast_2d(sample_rx)[0, :]

    res = correlate(sample_rx[::sps], symbol_tx)

    index = np.argmax(np.abs(res))

    out = np.roll(sample_rx, sps * (-index - 1 + symbol_tx.shape[0]))
    return out


def cd_compensation(signal: Signal, spans, inplace=False):
    '''

    This function is used for chromatic dispersion compensation in frequency domain.
    The signal is Signal object, and a new sample is created from property data_sample_in_fiber

    :param signal: Signal object
    :param spans: Span object, the signal's has propagated through these spans
    :param inplace: if True, the compensated sample will replace the original sample in signal,or new ndarray will be r
    eturned

    :return: if inplace is True, the signal object will be returned; if false the ndarray will be returned
    '''
    center_wavelength = signal.center_wave_length
    freq_vector = fftfreq(signal[0, :].shape[0], 1 / signal.fs_in_fiber)
    omeg_vector = 2 * np.pi * freq_vector

    sample = np.zeros_like(signal[:])

    for i in range(signal.shape[0]):
        sample[i, :] = signal[i, :]

    if not isinstance(spans, list):
        spans = [spans]

    for span in spans:
        beta2 = -span.beta2(center_wavelength)
        dispersion = (-1j / 2) * beta2 * omeg_vector ** 2 * span.length
        for pol_index in range(sample.shape[0]):
            sample[pol_index, :] = ifft(fft(sample[pol_index, :]) * np.exp(dispersion))

    if inplace:
        signal[:] = sample
        return signal
    else:
        return sample


#     return const[np.argmin(distance)]
def superscalar(symbol_in, training_symbol, block_length, pilot_number, constl, g,filter_n=20):
    # delete symbols to assure the symbol can be divided into adjecent channels
    symbol_in = np.atleast_2d(symbol_in)
    training_symbol =np.atleast_2d(training_symbol)
    constl = np.atleast_2d(constl)
    assert symbol_in.shape[0]==1
    assert training_symbol.shape[0]==1
    assert constl.shape[0]==1
    divided_symbols, divided_training_symbols = __initialzie_superscalar(symbol_in, training_symbol, block_length)
    angle = __initialzie_pll(divided_symbols, divided_training_symbols, pilot_number)

    divided_symbols = divided_symbols * np.exp(-1j * angle)
    divided_symbols = first_order_pll(divided_symbols, (constl), g)
    divided_symbols[0::2, :] = divided_symbols[0::2, ::-1]
    divided_symbols = divided_symbols.reshape((1, -1))
    # ml
    decision_symbols = np.zeros(divided_symbols.shape[1],dtype=np.complex)
    exp_decision(divided_symbols[0,:],constl[0,:],decision_symbols)

    hphase_ml = symbol_in[0,:len(decision_symbols)]/decision_symbols
    hphase_ml = np.atleast_2d(hphase_ml)
    h = np.ones((1,2*filter_n+1))
    hphase_ml = lfilter(h[0,:],1,hphase_ml)
    hphase_ml = np.roll(hphase_ml,-filter_n,axis=1)
    phase_ml = np.angle(hphase_ml)
    divided_symbols = symbol_in[:,:len(decision_symbols)] * np.exp(-1j*phase_ml)
    #ml completed
    divided_training_symbols[0::2, :] = divided_training_symbols[0::2, ::-1]
    divided_training_symbols = divided_training_symbols.reshape((1, -1))
    # scatterplot(divided_symbols,False,'pyqt')

    # filrst order pll

    return divided_symbols, divided_training_symbols


@numba.jit(nopython=True, parallel=True)
def first_order_pll(divided_symbols, constl, g):
    constl = np.atleast_2d(constl)
    phase = np.zeros((divided_symbols.shape[0], divided_symbols.shape[1]))
    for i in prange(divided_symbols.shape[0]):
        signal = divided_symbols[i, :]
        each_error = phase[i, :]
        for point_index, point in enumerate(signal):
            if point_index == 0:
                point = point * np.exp(-1j * 0)
            else:
                point = point * np.exp(-1j * each_error[point_index - 1])
            point_decision = decision(point, constl)
            signal[point_index] = point
            point_decision_conj = np.conj(point_decision)
            angle_difference = np.angle(point * point_decision_conj)

            if point_index > 0:
                each_error[point_index] = angle_difference * g + each_error[point_index - 1]
            else:
                each_error[point_index] = angle_difference * g

    return divided_symbols


def __initialzie_pll(divided_symbols, divided_training_symbols, pilot_number):
    '''
        There are pilot_number symbols of each row,the two adjecnt channel use the same phase,because they are simillar

    '''
    # get pilot symbol
    pilot_signal = divided_symbols[:, : pilot_number]
    pilot_traing_symbol = divided_training_symbols[:, :pilot_number]

    angle = (pilot_signal / pilot_traing_symbol)
    angle = angle.flatten()
    angle = angle.reshape(-1, 2 * pilot_number)
    angle = np.sum(angle, axis=1, keepdims=True)
    angle = np.angle(angle)

    angle_temp = np.zeros((angle.shape[0] * 2, angle.shape[1]), dtype=np.float)
    angle_temp[0::2, :] = angle
    angle_temp[1::2, :] = angle_temp[0::2, :]
    return angle_temp


def __initialzie_superscalar(symbol_in, training_symbol, block_length):
    # delete symbols to assure the symbol can be divided into adjecent channels
    symbol_in = np.atleast_2d(symbol_in)
    training_symbol = np.atleast_2d(training_symbol)
    assert symbol_in.shape[0] == 1
    symbol_length = len(symbol_in[0, :])
    assert divmod(block_length, 2)[1] == 0

    if divmod(symbol_length, 2)[1] != 0:
        # temp_symbol = np.zeros((symbol_in.shape[0], symbol_in.shape[1] - 1), dtype=np.complex)
        # temp_training_symbol = np.zeros((training_symbol.shape[0], training_symbol.shape[1] - 1), dtype=np.complex)
        temp_symbol = symbol_in[:, :-1]
        temp_training_symbol = training_symbol[:, :-1]
    else:
        temp_symbol = symbol_in
        temp_training_symbol = training_symbol

    # divide into channels
    channel_number = int(len(temp_symbol[0, :]) / block_length)
    if divmod(channel_number, 2)[1] == 1:
        channel_number = channel_number - 1
    divided_symbols = np.zeros((channel_number, block_length), dtype=np.complex)
    divided_training_symbols = np.zeros((channel_number, block_length), dtype=np.complex)
    for cnt in range(channel_number):
        divided_symbols[cnt, :] = temp_symbol[0, cnt * block_length:(cnt + 1) * block_length]
        divided_training_symbols[cnt, :] = temp_training_symbol[0, cnt * block_length:(cnt + 1) * block_length]
        if divmod(cnt, 2)[1] == 0:
            divided_symbols[cnt, :] = divided_symbols[cnt, ::-1]
            divided_training_symbols[cnt, :] = divided_training_symbols[cnt, ::-1]
    #             print(divided_training_symbols.shape)
    # First Order PLL

    return divided_symbols, divided_training_symbols


def dual_frequency_lms_equalizer_block(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001, niter=3):
    pass


@numba.jit(cache=True)
def dual_pol_time_domain_lms_equalizer_pll(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001,
                                           niter=3, g=0.01, symbol_length=65536):
    weight = __dual_pol_init_lms_weight(ntaps)  # xx xy yx yy
    data_sample_in_fiber = signal[:]
    samplex = segment_axis(data_sample_in_fiber[0, :], ntaps, ntaps - sps)
    sampley = segment_axis(data_sample_in_fiber[1, :], ntaps, ntaps - sps)
    if train_symbol is not None:
        xsymbol = train_symbol[0, ntaps // 2 // sps:]
        ysymbol = train_symbol[1, ntaps // 2 // sps:]
        # xsymbol = np.roll(xsymbol,)

    xsymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    ysymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    errorsx = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    errorsy = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    pll_phase = np.zeros((2, samplex.shape[0]), dtype=np.float64)
    # pll_error = np.zeros((2, samplex.shape[0]), dtype=np.complex128)

    hxx = np.zeros((1, samplex.shape[0]), dtype=np.float64)
    hxy = np.zeros((1, samplex.shape[0]), dtype=np.float64)
    hyx = np.zeros((1, samplex.shape[0]), dtype=np.float64)
    hyy = np.zeros((1, samplex.shape[0]), dtype=np.float64)

    for j in range(niter):
        for i in range(samplex.shape[0]):
            xout = np.sum(samplex[i, ::-1] * weight[0][0]) + np.sum(sampley[i, ::-1] * weight[1][0])
            yout = np.sum(samplex[i, ::-1] * weight[2][0]) + np.sum(sampley[i, ::-1] * weight[3][0])
            if i > 0:
                xout_cpr = xout * np.exp(-1j * pll_phase[0, i - 1])
                yout_cpr = yout * np.exp(-1j * pll_phase[1, i - 1])
            else:
                xout_cpr = xout * np.exp(-1j * 0)
                yout_cpr = yout * np.exp(-1j * 0)

            if train_symbol is not None:
                errorx = __calc_error_train(xout_cpr, symbol=xsymbol[i])
                errory = __calc_error_train(yout_cpr, symbol=ysymbol[i])
                xout_cpr_decision = xsymbol[i]
                yout_cpr_decision = ysymbol[i]
            else:
                assert constl is not None
                # assert constl.ndim ==1

                if constl.ndim != 2:
                    raise Exception("constl must be 2d")
                errorx = __calc_error_dd(xout_cpr, constl)
                errory = __calc_error_dd(yout_cpr, constl)
                xout_cpr_decision = decision(xout_cpr,constl)
                yout_cpr_decision = decision(yout_cpr,constl)

            angle_difference_x = np.angle((xout_cpr * np.conj(xout_cpr_decision)))
            angle_difference_y = np.angle((yout_cpr * np.conj(yout_cpr_decision)))
            if i == 0:
                pll_phase[0, i] = g * angle_difference_x
                pll_phase[1, i] = g * angle_difference_y
                pll_angle_x = 0
                pll_angle_y = 0
            else:
                pll_phase[0, i] = g * angle_difference_x + pll_phase[0, i - 1]
                pll_phase[1, i] = g * angle_difference_y + pll_phase[1, i - 1]
                pll_angle_x = pll_phase[0, i - 1]
                pll_angle_y = pll_phase[1, i - 1]

            weight[0][0] = weight[0][0] + mu * errorx * np.conj(samplex[i, ::-1] * np.exp(-1j * pll_angle_x))
            weight[1][0] = weight[1][0] + mu * errorx * np.conj(sampley[i, ::-1] * np.exp(-1j * pll_angle_y))

            weight[2][0] = weight[2][0] + mu * errory * np.conj(samplex[i, ::-1] * np.exp(-1j * pll_angle_x))
            weight[3][0] = weight[3][0] + mu * errory * np.conj(sampley[i, ::-1] * np.exp(-1j * pll_angle_y))

            if j == niter - 1:
                xsymbols[0, i] = xout_cpr
                ysymbols[0, i] = yout_cpr
                hxx[0, i] = weight[0][0, ntaps // 2 + 1].real
                hxy[0, i] = weight[1][0, ntaps // 2 + 1].real

                hyx[0, i] = weight[2][0, ntaps // 2 + 1].real

                hyy[0, i] = weight[3][0, ntaps // 2 + 1].real

            errorsx[0, i] = np.abs(errorx)
            errorsy[0, i] = np.abs(errory)

    return xsymbols[0, :], ysymbols[0, :], weight, errorsx, errorsy, (hxx, hxy, hyx, hyy)


@numba.jit(cache=True)
def dual_pol_time_domain_lms_equalizer(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001,
                                       niter=3, symbol_length=65536):
    '''

    :param signal:         signal to be equalized, a numpy array, 2d-array,each row represent a polarizaiton
    :param ntaps:          the tap of the filter
    :param sps:            sample per symbol
    :param constl:         the constellation, used in dd
    :param train_symbol:   symbol to train
    :param mu:             the learning rate, default to 0.001
    :param niter:          the number of equlizeing operation
    :return: equalized symbols, errors
    '''
    data_sample_in_fiber = signal[:]
    samplex = segment_axis(data_sample_in_fiber[0, :], ntaps, ntaps - sps)
    sampley = segment_axis(data_sample_in_fiber[1, :], ntaps, ntaps - sps)

    if train_symbol is not None:
        xsymbol = train_symbol[0, ntaps // 2 // sps:]
        ysymbol = train_symbol[1, ntaps // 2 // sps:]
        # xsymbol = np.roll(xsymbol,)

    weight = __dual_pol_init_lms_weight(ntaps)  # xx xy yx yy
    xsymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    ysymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    errorsx = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    errorsy = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)

    # hxx = np.zeros((1, samplex.shape[0]))
    # hxy = np.zeros((1, samplex.shape[0]))
    # hyx = np.zeros((1, samplex.shape[0]))
    # hyy = np.zeros((1, samplex.shape[0]))

    for j in range(niter):
        for i in range(samplex.shape[0]):
            xout = np.sum(samplex[i, ::-1] * weight[0][0]) + np.sum(sampley[i, ::-1] * weight[1][0])
            yout = np.sum(samplex[i, ::-1] * weight[2][0]) + np.sum(sampley[i, ::-1] * weight[3][0])

            if train_symbol is not None and j == 0:
                errorx = __calc_error_train(xout, symbol=xsymbol[i])
                errory = __calc_error_train(yout, symbol=ysymbol[i])
            else:
                assert constl is not None
                # assert constl.ndim ==1

                if constl.ndim != 2:
                    raise Exception("constl must be 2d")

                errorx = __calc_error_dd(xout, constl)
                errory = __calc_error_dd(yout, constl)

            weight[0][0] = weight[0][0] + mu * errorx * np.conj(samplex[i, ::-1])
            weight[1][0] = weight[1][0] + mu * errorx * np.conj(sampley[i, ::-1])

            weight[2][0] = weight[2][0] + mu * errory * np.conj(samplex[i, ::-1])
            weight[3][0] = weight[3][0] + mu * errory * np.conj(sampley[i, ::-1])

            if j == niter - 1:
                xsymbols[0, i] = xout
                ysymbols[0, i] = yout
            errorsx[0, i] = np.abs(errorx)
            errorsy[0, i] = np.abs(errory)

    return xsymbols[0,:], ysymbols[0,:], weight, errorsx, errorsy


@numba.jit(cache=True)
def __calc_error_dd(xout, constl):
    '''
        only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
        xout: a symbol to be decision,one element
        constl: must be 2d array
    '''
    symbol = decision(xout, constl)
    return symbol - xout


@numba.jit(cache=True)
def __calc_error_train(xout, symbol):
    '''
        only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
        xout: symbol in ,one element
        symbol: True symbol,one element
    '''
    error = symbol - xout
    return error


@numba.jit(cache=True)
def __dual_pol_init_lms_weight(ntaps):
    '''
         only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
         implement filter taps initialization

        :param ntaps:
        :return:
    '''
    # hxx = np.zeros((1, ntaps), dtype=np.complex128)
    #
    # hxy = np.zeros((1, ntaps), dtype=np.complex128)
    # hyx = np.zeros((1, ntaps), dtype=np.complex128)
    # hyy = np.zeros((1, ntaps), dtype=np.complex128)
    #
    # hxx[0, ntaps // 2] = 1
    # hyy[0, ntaps // 2] = 1

    hxx = np.zeros((1, ntaps), dtype=np.complex128)

    hxy = np.zeros((1, ntaps), dtype=np.complex128)
    hyx = np.zeros((1, ntaps), dtype=np.complex128)
    hyy = np.zeros((1, ntaps), dtype=np.complex128)

    hxx[0, ntaps // 2] = 1
    hyy[0, ntaps // 2] = 1
    return (hxx, hxy, hyx, hyy)


def demap_to_msg(rx_symbols, order, do_normal=True):
    '''
    :param rx_symbols:1d array or 2d array if 2d the shape[0] must be 1
    :param do_normal: if True the rx_symbols will be normalized to 1
    :return: 2d array msg, the shape[0] is 1
    '''

    from scipy.io import loadmat

    base_path = os.path.abspath(__file__)
    base_path = os.path.dirname(os.path.dirname(base_path))

    if order == 8:
        raise NotImplementedError('Not implemented yet')

    constl = cal_symbols_qam(order) / np.sqrt(cal_scaling_factor_qam(order))
    rx_symbols = np.atleast_2d(rx_symbols)
    assert rx_symbols.shape[0] == 1
    rx_symbols = rx_symbols[0]

    qam_data = loadmat(f'{base_path}/qamdata/{order}qam.mat')['x'][0]
    msg = np.zeros((1, rx_symbols.shape[0]))
    if do_normal:
        rx_symbols = rx_symbols / np.sqrt(np.mean(rx_symbols.imag ** 2 + rx_symbols.real ** 2))

    for index, symbol in enumerate(rx_symbols):
        decision_symbol = decision(symbol, constl=np.atleast_2d(constl))

        choose = np.abs(decision_symbol - qam_data) < np.spacing(1)
        #         print(choose)
        #         print(np.nonzero(choose)[0])
        msg[0, index] = (np.nonzero(choose)[0])
    #         print(msg[0,index])
    #         break

    return msg.astype(np.int)[0, :]


def main():
    import matplotlib.pyplot as plt
    from scipy.io import loadmat

    cc = loadmat('test2.mat')
    tx = cc['txsymbols']
    sam = cc['rxSignalIn']

    import time

    now = time.time()

    constl = cal_symbols_qam(16)
    constl = constl / np.sqrt(cal_scaling_factor_qam(16))
    constl = np.atleast_2d(constl)
    xsym, ysym, wer, errorsx, errorsy = dual_pol_time_domain_lms_equalizer(sam[:], 13, 2, constl=constl,
                                                                           train_symbol=tx,
                                                                           mu=0.001, niter=3)
    plt.plot(errorsx)
    print(time.time() - now)


if __name__ == '__main__':
    main()



import arrayfire as af
import numpy as np
from scipy.fftpack import fftfreq
from Ocspy.Base.SignalInterface import Signal
import math
def symmetric_ssfm(signal:Signal,setting:dict):
    '''
        :param:Signal a signal object
        :param:Setting: a dict contians the parameters of the fiber
        :Unit:
            alpha_lin: 1/km
            beta_2: s**2/km
            gamma: 1/w/km
            step_length:km
            fiber_length:km
    '''
    alpha_lin = setting['alpha_lin']
    beta2 = setting['beta2']
    gamma = setting['gamma']
    step_length = setting['step_length']
    fiber_length = setting['fiber_length']

    nsections = int(fiber_length / step_length)
    if nsections * step_length < fiber_length:
        last_step = fiber_length - nsections * step_length
    else:
        last_step = 0
    
    xpol = af.from_ndarray(signal[0,:])
    ypol = af.from_ndarray(signal[1, :])
    # xpol = samples[0,:]
    # ypol = samples[1,:]
    freq = fftfreq(len(xpol),1/signal.fs_in_fiber)
    omeg = 2*np.pi*freq
    omeg = af.from_ndarray(omeg)
    D = -1j/2 * beta2 * omeg**2
    step_length_eff = 1-math.exp(-alpha_lin*step_length)
    step_length_eff = step_length_eff / alpha_lin 
    for _ in range(nsections):
        xpol,ypol = linear_prop_arrayfire(xpol,ypol,step_length/2,D)
        xpol,ypol = nonlinear_prop(xpol,ypol,gamma,step_length_eff)
        xpol, ypol = linear_prop_arrayfire(xpol, ypol, step_length/2,D)
        xpol = xpol * math.exp(-alpha_lin/2 * step_length)
        ypol = ypol * math.exp(-alpha_lin/2 * step_length)
    if last_step:
        last_step_eff = 1-math.exp(-alpha_lin * last_step)
        last_step_eff = last_step_eff / alpha_lin
        xpol, ypol = linear_prop_arrayfire(xpol, ypol, last_step/2, D)
        xpol,ypol = nonlinear_prop(xpol,ypol,gamma,last_step_eff)
        xpol, ypol = linear_prop_arrayfire(xpol, ypol, last_step/2,D)
        xpol = xpol * math.exp(-alpha_lin/2 * last_step)
        ypol = ypol * math.exp(-alpha_lin/2 * last_step)

    xpol = xpol.to_ndarray()
    ypol = ypol.to_ndarray()
    xpol = np.asarray(xpol,order='C')
    ypol = np.asarray(ypol,order='C')
    return xpol,ypol

def nonlinear_prop(xpol,ypol,gamma,length):
    
    gamma = 1j*gamma* 8/9*length
    xpol = xpol * af.exp(gamma*(af.abs(xpol)**2 + af.abs(ypol)**2))
    ypol = ypol * af.exp(gamma*(af.abs(xpol)**2 + af.abs(ypol)**2))
    return xpol,ypol

def linear_prop_arrayfire(xpol,ypol,length,D):
    xpol_fft = af.fft(xpol)
    ypol_fft = af.fft(ypol)
    xpol_fft = xpol_fft * af.exp(D*length)
    ypol_fft = ypol_fft * af.exp(D*length)
    xpol =  af.ifft(xpol_fft)
    ypol =  af.ifft(ypol_fft)
    return xpol,ypol













import copy

import numpy as np
from scipy.io import loadmat
from scipy.constants import c
from typing import List
from ..tool.tool import freq2lamb
import os

base_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class Signal(object):
    '''
        Signal:
            This Signal is the base class for single carrier signal, the QamSignal inherit this class:
            the property of Signal:
                1. symbol_rate: GHZ
                2. mf: the modulation format
                3. symbol_length: the length of symbol
                4. signal_power: the launch power of signal,the waveform will be set to this signal,when the signal pass
                laser object
                5. sps: the sample per symbol for dsp
                6. sps in fiber： the sample per symbol for waveform in fiber,because of nonlinear effect of fiber, the sample
                rate shuould greater than 2 time highest frequence of signal.
                7.lam: the wavelength of the signal: [nm]

            @property

    '''

    def __init__(self, **kwargs):
        '''

        :param kwargs: 包含单信道信号的各项参数，包括：
            1. symbol_rate 符号速率 【Hz】
            2. mf 调制格式
            3. symbol length 符号长度
            4. signal_power  信号功率  【dbm】
            5. sps：         发端dsp的采样率
            6. sps_in_fiber: 光纤中采样率
            8. center_frequence: Hz

        property:
            symbol:  the symbol to be sent, can only be accessed but can not be set


        所有单位全部使用国际标准单位

        '''

        self._symbol = None
        self.symbol_rate = kwargs['symbol_rate']
        self.mf = kwargs['mf']
        self.symbol_length = kwargs['symbol_length']
        self.sps_in_fiber = kwargs['sps_in_fiber']
        self.sps = kwargs['sps']
        self.decision_symbol = None
        self.data_sample = None
        self.data_sample_in_fiber = None
        self.center_frequence = kwargs['frequence']

    @property
    def symbol(self):
        return self._symbol

    def inplace_set_signal_power(self,power,unit='dbm'):
        if unit =='dbm':
            pass
        else:
            power = power/10
            power = 10**power
            power = power/1000
        each_pol_power = power/self.pol_number
        self[:] = np.sqrt(each_pol_power)*self[:]

    def set_signal_power(self, power, unit="dbm"):
        if unit == 'dbm':
            power_lin = power / 10
            power_lin = 10 ** power_lin
            power_lin = power_lin / 1000
        else:
            power_lin = power
        self.inplace_normalize_power()
        for i in range(self.pol_number):
            self[i, :] = self[i, :] * np.sqrt(power_lin / self.pol_number)

    @property
    def sample_number_in_fiber(self):
        import warnings
        if self.sps_in_fiber * self.symbol_length != self.data_sample_in_fiber.shape[1]:
            warnings.warn(
                "the sps_in_fiber * symbol length not equal the sample number stored in sample_number_fiber")
            return self.data_sample_in_fiber.shape[1]

        return int(self.sps_in_fiber * self.symbol_length)

    @property
    def sample_number(self):
        return self.sps * self.symbol_length

    @property
    def center_wave_length(self):

        return c / self.center_frequence

    def tonumpy(self):
        '''

        :return: reference to self.data_sample_in_fiber
        if function not return a new ndarrya
        change it outside will change the data_sample_in_fiber property of the object
        '''
        data_sample_in_fiber = copy.deepcopy(self.data_sample_in_fiber)
        return data_sample_in_fiber

    def __getitem__(self, index):
        '''

        :param index: Slice Object
        :return: ndarray of the signal, change this ndarray will change the object

        '''
        assert self.data_sample_in_fiber is not None

        return self.data_sample_in_fiber[index]

    def __setitem__(self, index, value):
        '''

        :param   index: slice Object
        :param   value: the value that to be set,must be ndarray
        :return: None
        '''
        if self.data_sample_in_fiber is None:
            self.data_sample_in_fiber = np.atleast_2d(value)
        else:

            self.data_sample_in_fiber[index] = value

    def measure_power_in_fiber(self, unit='w'):
        '''

        :return: signal power in the fiber, in unit [W]
        '''
        sample_in_fiber = np.atleast_2d(self.data_sample_in_fiber)
        signal_power = 0
        for i in range(sample_in_fiber.shape[0]):
            signal_power += np.mean(np.abs(sample_in_fiber[i, :]) ** 2)
        if unit == "w":
            return signal_power
        else:
            return 10 * np.log10(signal_power * 1000)

    @property
    def fs(self):
        if self.sps is None:
            return 0
        return self.symbol_rate * self.sps

    @property
    def fs_in_fiber(self):
        return self.symbol_rate * self.sps_in_fiber

    def inplace_normalize_power(self):
        '''
            in place operation
        :return:
        '''
        self[:] = self.normalize_power()

    def remove_dc(self):
        for i in range(self.pol_number):
            self[i, :] = self[i, :] - np.mean(self[i, :])

    @property
    def pol_number(self):
        return self.data_sample_in_fiber.shape[0]

    def inplace_normalize_dsp_sample(self):
        temp = self.normalize_dsp_sample()
        self.data_sample = temp

    def normalize_dsp_sample(self):
        temp = np.zeros_like(self.data_sample)
        for i in range(self.pol_number):
            temp[i, :] = self.data_sample[i, :] / \
                np.sqrt(np.mean(np.abs(self.data_sample[i, :]) ** 2))
        return temp

    def normalize_power(self):
        '''

        new ndarray will be returned , the signal object itself is not changed
        :param signal:
        :return: ndarray and signal object will not be changed

        '''
        temp = np.zeros_like(self.data_sample_in_fiber)
        for i in range(self.data_sample_in_fiber.shape[0]):
            temp[i, :] = self.data_sample_in_fiber[i, :] / np.sqrt(
                np.mean(np.abs(self.data_sample_in_fiber[i, :]) ** 2))

        return temp

    def upsample(self, symbol_x, sps):
        '''

        :param symbol_x: 1d array
        :param sps: sample per symbol
        :return: 2-d array after inserting zeroes between symbols
        '''
        assert symbol_x.ndim == 1
        symbol_x.shape = -1, 1
        symbol_x = np.tile(symbol_x, (1, sps))
        symbol_x[:, 1:] = 0
        symbol_x.shape = 1, -1
        return symbol_x

    def __str__(self):
        try:
            string = f'\n\tSymbol rate:{self.symbol_rate / 1e9}[Ghz]\n\tfs_in_fiber:{self.fs_in_fiber / 1e9}[Ghz]\n\t' \
                f'signal_power_in_fiber:{self.measure_power_in_fiber()} [W]\n\t' \
                f'signal_power_in_fiber:{self.measure_power_in_fiber("dbm")} [dbm]\n\t' \
                f'wave_length is {self.center_wave_length * 1e9} [nm]\n\t' \
                f'center_frequence is {self.center_frequence * 1e-12}[Thz]\n\t'

        except Exception as e:
            string = f'\n\tSymbol rate:{self.symbol_rate / 1e9}[Ghz]\n\tfs_in_fiber:{self.fs_in_fiber / 1e9}[Ghz]\n\t' \
                f'wave_length is {self.center_wave_length * 1e9} [nm]\n\t' \
                f'center_frequence is {self.center_frequence * 1e-12}[Thz]\n\t'
        return string

    def __repr__(self):

        return self.__str__()

    @property
    def shape(self):
        return self.data_sample_in_fiber.shape


class SignalFromNumpyArray(Signal):

    def __init__(self, array_sample, symbol_rate, mf, symbol_length, sps_in_fiber, sps, **kwargs):
        '''

        :param array_sample: nd arrary
        :param symbol_rate: [Hz]
        :param mf: modulation format
        :param symbol_length: the length of transimitted symbol
        :param sps_in_fiber: the samples per symbol in fiber
        :param sps: the sampes per symbol in DSP

        This function will read samples from array_sample, the array_sample should be propogated in fiber

        The center frequence is the center frequence, and the absolute frequence is from left to right in wdm comb

        The relative frequence i.e base band equivalent is set a property
        '''
        super(SignalFromNumpyArray, self).__init__(
            **dict(symbol_rate=symbol_rate, mf=mf, symbol_length=symbol_length, sps_in_fiber=sps_in_fiber, sps=sps))
        self.data_sample_in_fiber = array_sample

        if 'tx_symbol' in kwargs:
            self.tx_symbol = kwargs['tx_symbol']
        if 'rx_symbol' in kwargs:
            self.rx_symbol = kwargs['rx_symbol']

        if 'center_frequence' in kwargs:
            self.center_frequence = kwargs['center_frequence']

        if 'aboslute_frequence' in kwargs:
            self.absolute_frequence = kwargs['absolute_frequence']

    @property
    def relative_frequence(self):
        if hasattr(self, 'center_frequence') and hasattr(self, 'absolute_frequence'):
            return self.absolute_frequence - self.center_frequence
        else:
            raise Exception(
                'The frequence not provided enough, please ensure the center_frequency and the absolute_frequenc are all provided')


class WdmSignal(object):
    '''
        WdmSignal Class
    '''

    def __init__(self, signals: List[Signal], grid_size, multiplexed_filed):
        '''

        :param signals: list of signal
        :param grid_size: [Hz],if conventional grid_size,this will be a number or a list of same value
        if the elestical wdm, the grid size of each channel is different.
        :param center_frequence: Hz

        The WdmSignal class should not be used outside usually, in mux function, it will be created automatically.

        '''
        self.signals = signals
        self.grid_size = grid_size
        self.center_frequence = (
            np.min(self.absolute_frequences) + np.max(self.absolute_frequences)) / 2
        self.center_wave_length = freq2lamb(self.center_frequence)
        self.__multiplexed_filed = np.atleast_2d(multiplexed_filed)

    def __setitem__(self, slice, value):

        assert self.__multiplexed_filed is not None
        self.__multiplexed_filed[slice] = value

    def __getitem__(self, slice):
        assert self.__multiplexed_filed is not None
        return self.__multiplexed_filed[slice]

    # @property
    # def center_frequence(self):
    #     frequences = self.absolute_frequences()
    #     return (np.max(frequences)+np.min(frequences))/2

    @property
    def pol_number(self):
        return self.__multiplexed_filed.shape[0]

    @property
    def shape(self):
        return self.__multiplexed_filed.shape

    @property
    def fs_in_fiber(self):
        return self.signals[0].fs_in_fiber

    @property
    def sample_number_in_fiber(self):
        return self.__multiplexed_filed.shape[1]

    @property
    def lam(self):
        lambdas = []
        for signal in self.signals:
            lambdas.append(signal.center_wave_length)
        return lambdas

    @property
    def data_sample_in_fiber(self):
        return self.__multiplexed_filed

    @data_sample_in_fiber.setter
    def data_sample_in_fiber(self, value):
        self.__multiplexed_filed = value

    @property
    def mfs(self):
        mfs = []
        for sig in self.signals:
            mfs.append(sig.mf)
        return mfs

    @property
    def absolute_frequences(self):
        '''

        :return:
        '''
        return np.array([signal.center_frequence for signal in self.signals])

    @property
    def relative_frequences(self):
        '''

        :return: frequences centered in zero frequence , an ndarray
        '''
        return self.absolute_frequences - self.center_frequence

    def __str__(self):

        string = f"center_frequence is {self.center_frequence * 1e-12} [THZ]\t\n" \
            f"power is {self.measure_power_in_fiber()} [dbm] \t\n" \
            f"power is {self.measure_power_in_fiber(unit='w')} [w]\t\n"

        return string

    def measure_power_in_fiber(self, unit='dbm'):

        power = 0
        for i in range(self.pol_number):
            power += np.mean(np.abs(self[i, :]) ** 2)

        if unit == 'dbm':
            power = power * 1000
            power = 10 * np.log10(power)

        return power

class WdmSignalFromArray(object):

    def __init__(self,field,frequencys,fs_in_fiber,signal_under_study,signal_index):
        '''

        :param field:   WDM multiplexed fiedl
        :param center_freq: Hz will be caculated from frequencys
        :param frequencys: Hz each channel's frequencys, the passband frequence
        :param fs_in_fiber: Hz
        '''
        self.__filed = field
        self.fs_in_fiber = fs_in_fiber
        self.frequences = frequencys
        self.center_frequence = np.mean([np.max(frequencys),np.min(frequencys)])
        self.center_wave_length = freq2lamb(self.center_frequence) #m
        self.signal_under_study = signal_under_study
        self.signal_index = signal_index

    def __setitem__(self, slice, value):

        assert self.__filed is not None
        self.__filed[slice] = value

    def __getitem__(self, slice):

        assert self.__filed is not None
        return self.__filed[slice]

    @property
    def pol_number(self):
        return self.__filed.shape[0]

    @property
    def shape(self):
        return self.__filed.shape

    @property
    def sample_number_in_fiber(self):
        return self.__filed.shape[1]

    @property
    def lam(self):
        lambdas = []
        for freq in self.frequences:
            lambdas.append(freq2lamb(freq))
        return lambdas

    @property
    def data_sample_in_fiber(self):
        return self.__filed

    @data_sample_in_fiber.setter
    def data_sample_in_fiber(self, value):
        self.__filed = value

    @property
    def absolute_frequences(self):
        '''

        :return:
        '''
        return self.frequences

    @property
    def relative_frequences(self):
        '''

        :return: frequences centered in zero frequence , an ndarray
        '''
        return self.absolute_frequences - self.center_frequence

    def __str__(self):

        string = f"center_frequence is {self.center_frequence * 1e-12} [THZ]\t\n" \
            f"power is {self.measure_power_in_fiber()} [dbm] \t\n" \
            f"power is {self.measure_power_in_fiber(unit='w')} [w]\t\n"

        return string

    def measure_power_in_fiber(self, unit='dbm'):

        power = 0
        for i in range(self.pol_number):
            power += np.mean(np.abs(self[i, :]) ** 2)

        if unit == 'dbm':
            power = power * 1000 # convert to mw
            power = 10 * np.log10(power)

        return power

class QamSignal(Signal):

    def __init__(self, symbol_rate=35e9, mf="16-qam", symbol_length=2 ** 16, sps=2, sps_in_fiber=4,
                 is_dp=True, frequence=193.1e12, is_from_array=False, is_from_demux=False):
        '''

        :param symbol_rate: [hz]   符号速率
        :param mf: 调制格式,16-qam,32-qam,64-qam,qpsk
        :param signal_power: [dbm] 信号功率
        :param symbol_length:      符号长度
        :param sps: 发端dsp的过采样率
        :param sps_infiber: 在光纤中传输时的过采样率
        '''

        super().__init__(
            **dict(symbol_length=symbol_length, symbol_rate=symbol_rate, mf=mf, sps=sps,
                   sps_in_fiber=sps_in_fiber, frequence=frequence)
        )
        self.is_from_demux = is_from_demux
        if self.mf == 'qpsk':
            order = 4
        else:
            order = self.mf.split('-')[0]
            order = int(order)
        if not is_from_array and not self.is_from_demux:
            if is_dp:

                self.message = np.random.randint(
                    low=0, high=order, size=(2, self.symbol_length))
            else:
                self.message = np.random.randint(
                    low=0, high=order, size=(1, self.symbol_length))

            self.is_dp = is_dp
            # self.symbol = None

            self.init(order)

    def init(self, order):

        qam_data = (f'{base_path}/' + 'qamdata/' + str(order) + 'qam.mat')
        qam_data = loadmat(qam_data)['x']

        symbol = np.zeros(self.message.shape, dtype=np.complex)
        symbol = np.atleast_2d(symbol)
        for i in range(symbol.shape[0]):
            for msg in np.unique(self.message[i, :]):
                symbol[i, self.message[i, :] == msg] = qam_data[0, msg]

        self._symbol = symbol




from .tool import scatterplot
from .tool import save_signal
from .tool import load_signal
from .tool import spectrum_analyzer
from .tool import lamb2freq
from .tool import freq2lamb
from .tool import downsample


# -*- coding: utf-8 -*-
#  This file is part of QAMpy.
#
#  QAMpy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Foobar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with QAMpy.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 Jochen Schröder, Mikael Mazur

# These two functions from QamPy library


import warnings
from numba.types import float64
import numpy as np
import numba

N = np

""" a number of convenience functions"""


def normal_sample(samples):
    '''

    :param samples: 1d array or 2d array
    :return: normalized samples
    '''
    assert isinstance(samples,np.ndarray)
    ndim = samples.ndim
    samples = np.atleast_2d(samples)
    samples_new = np.zeros_like(samples,dtype=samples.dtype)
    cnt = 0
    for row in samples:
        row = row / np.sqrt(np.mean(row.real ** 2 + row.imag ** 2))
        samples_new[cnt,:] = row
        cnt = cnt+1
    if ndim ==1:
        samples_new = samples_new[0,:]
    return samples_new


@numba.jit(cache=True)
def segment_axis(a, length, overlap, mode='cut', append_to_end=0):
    """
        Generate a new array that chops the given array along the given axis into
        overlapping frames.

        example:
        >>> segment_axis(arange(10), 4, 2)
        array([[0, 1, 2, 3],
               [2, 3, 4, 5],
               [4, 5, 6, 7],
               [6, 7, 8, 9]])

        arguments:
        a       The array to segment must be 1d-array
        length  The length of each frame
        overlap The number of array elements by which the frames should overlap

        end     What to do with the last frame, if the array is not evenly
                divisible into pieces. Options are:

                'cut'   Simply discard the extra values
                'pad'   Pad with a constant value

        append_to_end:    The value to use for end='pad'

        a new array will be returned.

    """

    if a.ndim !=1:
        raise Exception("Error, input array must be 1d")
    if overlap > length:
        raise Exception("overlap cannot exceed the whole length")

    stride = length - overlap
    row = 1
    total_number = length
    while True:
        total_number = total_number + stride
        if total_number > len(a):
            break
        else:
            row = row + 1

    # 一共要分成row行
    if total_number > len(a):
        if mode == 'cut':
            b = np.zeros((row, length), dtype=np.complex128)
            is_append_to_end = False
        else:
            b = np.zeros((row + 1, length), dtype=np.complex128)
            is_append_to_end = True
    else:
        b = np.zeros((row, length), dtype=np.complex128)
        is_append_to_end = False

    index = 0
    for i in range(row):
        b[i, :] = a[index:index + length]
        index = index + stride

    if is_append_to_end:
        last = a[index:]

        b[row, 0:len(last)] = last
        b[row, len(last):] = append_to_end

    return b


def bin2gray(value):
    """
    Convert a binary value to an gray coded value see _[1]. This also works for arrays.
    ..[1] https://en.wikipedia.org/wiki/Gray_code#Constructing_an_n-bit_Gray_code
    """
    return value ^ (value >> 1)



@numba.guvectorize([(numba.types.complex128[:],numba.types.complex128[:],numba.types.complex128[:])], '(n),(m)->(n)',nopython=True)
def exp_decision(symbol,const,res):
    for index,sym in enumerate(symbol):
        # distance = sym-const
        distance = np.abs(sym-const)
        res[index] = const[np.argmin(distance)]




@numba.jit(cache=True)
def decision(symbol, constl):
    '''
        constl must be 2d
    '''
    if constl.ndim !=2:
        raise Exception("constl input to decision must be 2d")
    distance = np.abs(constl[0] - symbol)
    decision = constl[0, np.argmin(distance)]

    return decision


def cal_symbols_qam(M):
    """
    Generate the symbols on the constellation diagram for M-QAM
    """
    if np.log2(M) % 2 > 0.5:
        return cal_symbols_cross_qam(M)
    else:
        return cal_symbols_square_qam(M)


def cal_symbols_cross_qam(M):
    """
    Generate the symbols on the constellation diagram for non-square (cross) M-QAM
    """
    N = (np.log2(M) - 1) / 2
    s = 2 ** (N - 1)
    rect = np.mgrid[-(2 ** (N + 1) - 1):2 ** (N + 1) - 1:1.j * 2 ** (N + 1), -(
            2 ** N - 1):2 ** N - 1:1.j * 2 ** N]
    qam = rect[0] + 1.j * rect[1]
    idx1 = np.where((abs(qam.real) > 3 * s) & (abs(qam.imag) > s))
    idx2 = np.where((abs(qam.real) > 3 * s) & (abs(qam.imag) <= s))
    qam[idx1] = np.sign(qam[idx1].real) * (
            abs(qam[idx1].real) - 2 * s) + 1.j * (np.sign(qam[idx1].imag) *
                                                  (4 * s - abs(qam[idx1].imag)))
    qam[idx2] = np.sign(qam[idx2].real) * (
            4 * s - abs(qam[idx2].real)) + 1.j * (np.sign(qam[idx2].imag) *
                                                  (abs(qam[idx2].imag) + 2 * s))
    return qam.flatten()


def cal_symbols_square_qam(M):
    """
    Generate the symbols on the constellation diagram for square M-QAM
    """
    qam = np.mgrid[-(2 * np.sqrt(M) / 2 - 1):2 * np.sqrt(
        M) / 2 - 1:1.j * np.sqrt(M), -(2 * np.sqrt(M) / 2 - 1):2 * np.sqrt(M) /
                                                               2 - 1:1.j * np.sqrt(M)]
    return (qam[0] + 1.j * qam[1]).flatten()


def cal_scaling_factor_qam(M):
    """
    Calculate the scaling factor for normalising MQAM symbols to 1 average Power
    """
    bits = np.log2(M)
    if not bits % 2:
        scale = 2 / 3 * (M - 1)
    else:
        symbols = cal_symbols_qam(M)
        scale = (abs(symbols) ** 2).mean()
    return scale


def get_power(x):
    return np.mean(x.real ** 2 + x.imag ** 2)



import pickle
import zlib

import numpy as np
import plotly.graph_objs as go
from plotly import tools
from scipy.constants import c
from scipy.signal import welch
import pyqtgraph as pg
from pyqtgraph.Qt import QtGui,QtCore
def lamb2freq(lam):
    '''

    :param lam: wavelength [m]
    :return: frequence [Hz]
    '''
    return c / lam


def freq2lamb(freq):
    '''

    :param freq: frequence [Hz]
    :return: lambda:[m]
    '''
    return c / freq


def downsample(signal, sps):
    '''
    downsample along row

    can receive siganl object or ndarray
    :param signal: Signal Object or ndarray
    :param sps:
    :return: signal
    '''

    if not isinstance(signal, np.ndarray):
        sample = signal.data_sample_in_fiber
    else:
        sample = np.atleast_2d(signal)

    length = divmod(sample.shape[1], sps)

    if length[1] != 0:
        raise ("after downsample the sample number should be integer")
    after_process = np.zeros((sample.shape[0], length[0]), dtype=sample.dtype)

    for i in range(sample.shape[0]):
        after_process[i, :] = sample[i, ::sps]

    return after_process


def scatterplot_colorful(signal, down_sample=True):
    pass


def scatterplot(signal, down_sample=True, backend='mplt', visdom_env="scatterplot", server='http://localhost',
                port=8097, head=1024, sps=None):
    '''

    :param signal:      signal object or ndarray
    :param down_sample: if down_sample is true , signal must be signal object
    :param backend:     can choose mplt or pyqt
    :return:
    '''

    if down_sample:
        if hasattr(signal, 'sps_in_fiber'):
            sps = signal.sps_in_fiber
        else:
            assert sps is not None

        symbol = downsample(signal, sps)
        symbol = np.atleast_2d(symbol)
        symbol = symbol[:, head:-head]
    else:
        if hasattr(signal, 'decision_symbol'):
            symbol = signal.decision_symbol
            symbol = np.atleast_2d(symbol)
        else:
            symbol = np.atleast_2d(signal)
            symbol = symbol[:, head:-head]
    pol_number = symbol.shape[0]
    if backend == 'mplt':
        import matplotlib.pyplot as plt
        if pol_number ==2:
            plt.figure(figsize=(12, 6))
        if pol_number ==1:
            plt.figure(figsize=(6, 6))

        for i in range(pol_number):
            plt.subplot(1, pol_number, i + 1,aspect='equal')
            ibranch = symbol[i, :].real
            qbranch = symbol[i, :].imag
            plt.scatter(ibranch, qbranch, marker='o', color='b')
        plt.show()

    if backend == 'visdom':
        import visdom
        viz = visdom.Visdom(env=visdom_env, server=server, port=port)
        traces = []
        for i in range(pol_number):
            trace = go.Scattergl(x=symbol[i, :].real, y=symbol[i, :].imag, mode='markers')
            traces.append(trace)
        fig = tools.make_subplots(rows=1, cols=pol_number, print_grid=False)

        for index, trace in enumerate(traces):
            fig.append_trace(trace, 1, index + 1)
        fig['layout'].update(height=600, width=800)
        for i in range(pol_number):
            fig['layout'][f'xaxis{i + 1}'].update(title='I-Branch')
            fig['layout'][f'yaxis{i + 1}'].update(title='Q-Branch')

            fig['layout'][f'xaxis{i + 1}']['zeroline'] = True
            fig['layout'][f'yaxis{i + 1}']['zeroline'] = True
            # fig['layout'][f'xaxis{i+1}']['range'] = [np.max(symbol[i,:].real)+0.02,np.min(symbol[i,:].real)-0.02]
            #
            # fig['layout'][f'yaxis{i+1}']['range'] = [np.max(symbol[i,:].imag)+0.02,np.min(symbol[i,:].imag)-0.02]
        viz.plotlyplot(fig)
    if backend == 'pyqt':
        __scatterplot_pyqt(symbol)


def __scatterplot_pyqt(sample):
    is_pol_demux = sample.shape[0] == 2
    if is_pol_demux:
        width = 1200
        height = 600
    else:
        width = 800
        height = 800
    app, mw, view = __pyqtgraph_init(width, height)
    mw.show()
    for i in range(sample.shape[0]):
        w1 = view.addPlot(row=0, col=i)

        s1 = pg.ScatterPlotItem(size=3, pen=pg.mkPen(color='b'), brush=pg.mkBrush(color='b'))
        sample_to_draw = sample[i, :]
        ibranch = sample_to_draw.real
        qbranch = sample_to_draw.imag
        s1.addPoints(x=ibranch, y=qbranch)
        w1.addItem(s1)
        w1.setLabel('left', "Q-Signal")
        w1.setLabel('bottom', "I-Signal")
    QtGui.QApplication.instance().exec_()


def __pyqtgraph_init(widht, height):
    app = QtGui.QApplication([])

    mw = QtGui.QMainWindow()
    mw.resize(widht, height)
    view = pg.GraphicsLayoutWidget()  ## GraphicsView with GraphicsLayout inserted by default
    mw.setCentralWidget(view)
    return app, mw, view


def spectrum_analyzer(signal, fs=None, backend='mplt', visdom_env='spectrum', server='http://localhost',
                      port=8097):
    '''

    :param signal: signal object or ndarray
    :return: None
    '''

    if isinstance(signal, np.ndarray):
        assert fs is not None
        sample = signal
    else:
        fs = signal.fs_in_fiber

        sample = signal[:]

    pol_number = sample.shape[0]
    if backend == 'mplt':
        import matplotlib.pyplot as plt
        plt.figure(figsize=(20, 6))
        for i in range(pol_number):
            plt.subplot(1, pol_number, i + 1)
            [f, pxx] = welch(sample[i, :], fs, nfft=2048, detrend=False, return_onesided=False)
            plt.plot(f / 1e9, 10 * np.log10(np.abs(pxx)))
            plt.xlabel('Frequency [GHZ]')
            plt.ylabel('Power Spectrem Density [db/Hz]')
        plt.show()
    if backend == 'visdom':
        import visdom
        viz = visdom.Visdom(env=visdom_env, server=server, port=port)
        traces = []

        for i in range(pol_number):
            [f, pxx] = welch(sample[i, :], fs, nfft=2048, detrend=False, return_onesided=False)
            trace = go.Scattergl(x=f, y=10 * np.log10(np.abs(pxx)), mode='lines', )
            traces.append(trace)
        fig = tools.make_subplots(rows=1, cols=pol_number, print_grid=False)

        for index, trace in enumerate(traces):
            fig.append_trace(trace, 1, index + 1)
        fig['layout'].update(height=400, width=800)
        for i in range(pol_number):
            fig['layout'][f'xaxis{i + 1}'].update(title='I-Branch')
            fig['layout'][f'yaxis{i + 1}'].update(title='Q-Branch')
        viz.plotlyplot(fig)


def eyedigram(signal):
    pass


def save_signal(fn, signal, lvl=4, **kwargs):
    """
    Save a signal object using zlib compression

    Parameters
    ----------
    fn : basestring
        filename
    signal : SignalObject
        the signal to save
    lvl : int, optional
        the compression to use for zlib
    """
    with open(fn, "wb") as fp:
        sc = zlib.compress(
            pickle.dumps({'signal': signal, 'simulation_information': kwargs}, protocol=pickle.HIGHEST_PROTOCOL),
            level=lvl)
        fp.write(sc)


def load_signal(fn):
    """
    Load a signal object from a zlib compressed pickle file.

    Parameters
    ----------
    fn : basestring
        filename of the file

    Returns
    -------
    sig : SignalObject
        The loaded signal object

    """
    with open(fn, "rb") as fp:
        s = zlib.decompress(fp.read())
        obj = pickle.loads(s)
    return obj



import numpy as np
from .dsp_tools import cal_symbols_qam
from .dsp_tools import cal_scaling_factor_qam
from .dsp_tools import normal_sample
import numba


def biterror(order, receive_msg, tx_msg):
    assert receive_msg.ndim == 1
    assert tx_msg.ndim == 1
    assert len(receive_msg) == len(tx_msg)
    bitnumber = int(np.log2(order))

    cnt = 0

    index = receive_msg != tx_msg
    to_find_recv = receive_msg[index]
    to_find_tx = tx_msg[index]

    for i in range(len(to_find_recv)):
        temp_recv = np.binary_repr(to_find_recv[i], bitnumber)
        temp_tx = np.binary_repr(to_find_tx[i], bitnumber)
        for index in range(len(temp_recv)):
            if temp_recv[index] != temp_tx[index]:
                cnt = cnt + 1
    return cnt / len(receive_msg) / bitnumber


def Q(biterr):
    pass


def evm(receive_symbol, tx_symbol):
    pass


def theory_ber_from_snr(M, is_pdm):
    pass


def snr2osnr(is_pdm):
    pass


def osnr2snr(is_pdm, signal_bandwidth):
    pass


def estimate_snr_using_tx(recive_symbol, tx_symbol, to_normalize=True, head=1024):
    recive_symbol = np.atleast_2d(recive_symbol)
    pol_number = recive_symbol.shape[0]

    tx_symbol = np.atleast_2d(tx_symbol)
    assert recive_symbol.shape == tx_symbol.shape

    recive_symbol = recive_symbol[:, head:-head]
    tx_symbol = tx_symbol[:, head:-head]

    snr = []

    noise_powers = []
    if to_normalize:
        for i in range(pol_number):
            recive_symbol[i, :] = normal_sample(recive_symbol[i, :])
            tx_symbol[i, :] = normal_sample(tx_symbol[i, :])
    else:
        print("please ensure the rx and tx are normalized")

    noise = recive_symbol - tx_symbol

    for i in range(pol_number):
        noise_power = noise[i, :].real ** 2 + noise[i, :].imag ** 2
        noise_power = np.mean(noise_power)

        noise_powers.append(noise_power)

        snr.append(10 * np.log10(1 / noise_power))

    if pol_number == 2:
        total_noise_power = np.sum(noise_powers)
        total_snr = 2 / total_noise_power

        total_snr = 10 * np.log10(total_snr)
        snr.append(total_snr)
        return snr[0], snr[1], snr[2]

    else:
        return snr[0]


def estimate_snr(E, M):
    """Calculate the signal to noise ratio SNR according to formula given in
    _[1]

    Parameters:
    ----------
    E   : array_like
      input field
    M:  : int
      order of the QAM constallation

    Returns:
    -------
    S0/N: : float
        linear SNR estimate

    References:
    ----------
    ...[1] Gao and Tepedelenlioglu in IEEE Trans in Signal Processing Vol 53,
    pg 865 (2005).

    """
    gamma = _cal_gamma(M)
    r2 = np.mean(abs(E) ** 2)
    r4 = np.mean(abs(E) ** 4)
    S1 = 1 - 2 * r2 ** 2 / r4 - np.sqrt(
        (2 - gamma) * (2 * r2 ** 4 / r4 ** 2 - r2 ** 2 / r4))
    S2 = gamma * r2 ** 2 / r4 - 1
    return S1 / S2


def _cal_gamma(M):
    """Calculate the gamma factor for SNR estimation."""
    A = abs(cal_symbols_qam(M)) / np.sqrt(cal_scaling_factor_qam(M))
    uniq, counts = np.unique(A, return_counts=True)
    return np.sum(uniq ** 4 * counts / M)
